#!python
# Copyright (c) 2000-2003 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

"""
    Komodo Regression Test Suite

    Usage:

        bk test [<options>...] [<tests>...]

    Options:
        -v, --verbose   run tests in verbose mode with output to stdout
        -q, --quiet     don't print anything except if a test fails
        -h, --help      print this text and exit
        -x <testname>, --exclude=<testname>
                        Exclude the named test from the set of tests to be
                        run.  This can be used multiple times to specify
                        multiple exclusions.
        -l, --list      Just list the test modules. Don't run them.
        --gui, --text   Just run the GUI or non-GUI tests. By default
                        both sets of tests are run.

    Smoke Options:
        --smoke             Log test results in Smoke.
        -s <server>, --server=<server>
                            Smoke server to use.
        --project-id=<id>   Smoke project id to use. Defaults to the
                            given Smoke server's project_id for
                            "komodo".
        --build-id=<id>     Smoke build id to use. Defaults to a "dev
                            build id" for the current machine and Komodo
                            "build tree" build.

    This will find all modules whose name is "test_*" in the test and
    other directories, and run them.  Various command line options
    provide additional facilities.

    If non-option arguments are present, they are names for tests to
    run.  Specifying particular tests to run or tests to exclude has the
    side-effect of skipping all the GUI tests.  If no test names are
    given, all tests are run.

    Examples:
        bk test                        # test everything
        bk test --gui                  # just run the GUI tests
        bk test --text                 # just run the text tests
        bk test koDocument             # run specific text tests
        bk test -x startup,koIndexer   # exclude some tests (skips GUI
                                       # tests too)
"""
#TODO:
#   - Look at the tests in these directories. They need work before
#     being included in the test suite.
#       src/debugger
#       src/debugger/korpc
#       src/invocations
#       src/python-sitelib/test
#           Actually, we don't want process.py's tests if process.py
#           gets a life of its own outside Komodo
#       src/scc/cvs/test
#           Need to work on cvslib's test to somehow deal with the
#           setup/teardown stuff needed for all those test modules.
#           Perhaps move them all to one test_cvslib.py, but that still
#           does not solve the setup/teardown problem.

import os
import sys
import getopt
import glob
import time
import types
import tempfile
import unittest
import pprint
import imp
import logging

import smokeclient



#---- exceptions

class TestError(Exception):
    pass



#---- globals

log = logging.getLogger("test")

gTestDirs = [
    # Add the relative path (relative to test.py) to each directory from
    # which to gather test_* modules.
    os.curdir,
]



#---- utility routines

def _getTests(dir):
    """Return a list of test names in the given directory."""
    testPyFiles = glob.glob(os.path.join(dir, "test_*.py"))
    modules = [os.path.splitext(os.path.basename(f))[0]
               for f in testPyFiles if f and f.endswith(".py")]

    packages = []
    for f in glob.glob(os.path.join(dir, "test_*")):
        if os.path.isdir(f) and "." not in f:
            if os.path.isfile(os.path.join(dir, f, "__init__.py")):
                packages.append(os.path.basename(f))

    return modules + packages


def _setUp():
    pass


def _tearDown():
    pass


def test(categories, modules, excludes=[], justList=0,
         # Smoke options:
         useSmoke=0, server=None, project_id=None, build_id=None):
    log.debug("test(categories=%r, modules=%r, excludes=%r, useSmoke=%r, "
              "server=%r, project_id=%r, build_id=%r)", categories,
              modules, excludes, useSmoke, server, project_id, build_id)

    # Trim '.py' from use supplied arguments. They might have gotten
    # there via shell expansion.  Also, allow the user to optionally
    # drop the standard "test_" prefix from given test names.
    for i in range(len(modules)):
        if modules[i].endswith(".py"):
            modules[i] = modules[i][:-3]
        if not modules[i].startswith("test_"):
            modules[i] = "test_"+modules[i]
    for i in range(len(excludes)):
        if excludes[i].endswith(".py"):
            excludes[i] = excludes[i][:-3]
        if not excludes[i].startswith("test_"):
            excludes[i] = "test_"+excludes[i]

    if "text" in categories:
        # Find all available tests.
        testmap = {} # mapping of test names to their directory
        for dir in gTestDirs:
            for name in _getTests(dir):
                if name in testmap:
                    raise TestError("test name collision: test '%s' existing "
                                    "in both '%s' and '%s'", name,
                                    testmap[name], dir)
                testmap[name] = dir
        
        # If user specified modules, then use only those.
        #XXX Perhaps should warn or error out if any or all of given test
        #    names do not exist?
        if modules:
            for name in testmap.keys():
                if name not in modules:
                    del testmap[name]

        # Drop specifically excluded modules.
        for name in excludes:
            if name in testmap:
                del testmap[name]

        if justList:
            pprint.pprint(testmap)
            return

        # Aggregate the TestSuite's from each module into one big one.
        suites = []
        for name, dir in testmap.items():
            file, filename, desc = imp.find_module(name, [dir])
            try:
                module = imp.load_module(name, file, filename, desc)
            finally:
                if file:
                    file.close()
            suite = getattr(module, "suite", None)
            if suite is not None:
                suites.append(suite())
            else:
                log.warn("module '%s' (in '%s') does not have a suite() method",
                         name, dir)
        suite = unittest.TestSuite(suites)

        # Run the suite.
        if useSmoke:
            from smokeclient.unittestutils import SmokeTestRunner
            runner = SmokeTestRunner(server=server, project_id=project_id,
                                     build_id=build_id)
        else:
            runner = unittest.TextTestRunner(sys.stdout, verbosity=2)
        result = runner.run(suite)

    if ("gui" in categories
        # Kind of a hack: if the user specified tests to run or exclude
        # and didn't *specifically* ask for GUI tests, then don't run
        # the GUI tests.
        and not ("text" in categories and (modules or excludes))
       ):
        if "text" in categories:
            time.sleep(3) # give the "text" tests time to settle down

        log.info("start Komodo Jaguar")
        import which
        assert which._version_ >= (1,0,0)
        import process
        argv = [which.which("komodo"),
                "-v"]   # open debugging window for now
        p = process.Process(argv)
        time.sleep(1) # give it time to "announce" that it is starting up

        log.info("run GUI test suite")
        import kocommandments as kc
        if useSmoke:
            args = ["--smoke", "--server", server]
            if project_id:
                args += ["--project-id", project_id]
            if build_id:
                args += ["--build-id", build_id]
        else:
            args = []
        kc.issue("testgui", args) #XXX Want this to be synchronous.

        #XXX Do we want to shutdown? If so, when?
        time.sleep(5)
        log.info("shutting down Komodo Jaguar")
        kc.issue("quit")



#---- mainline

def main(argv):
    logging.basicConfig()
    log.setLevel(logging.INFO)

    # parse options
    try:
        opts, modules = getopt.getopt(argv[1:], "hvqx:ls:",
            ["help", "verbose", "quiet", "exclude=", "list", "gui",
             "text", "smoke", "server=", "project-id=", "build-id="])
    except getopt.error, ex:
        log.error(str(ex))
        return 1
    excludes = []
    justList = 0
    useSmoke = 0
    server = None
    project_id = None
    build_id = None
    categories = ["gui", "text"]
    for opt, optarg in opts:
        if opt in ("-h", "--help"):
            print __doc__
            return 0
        elif opt in ("-v", "--verbose"):
            log.setLevel(logging.DEBUG)
        elif opt in ("-q", "--quiet"):
            log.setLevel(logging.WARN)
        elif opt in ("-x", "--exclude"):
            excludes += optarg.split(',')
        elif opt in ("-l", "--list"):
            justList = 1
        elif opt == "--gui":
            categories = ["gui"]
        elif opt == "--text":
            categories = ["text"]
        elif opt == "--smoke":
            useSmoke = 1
        elif opt in ("-s", "--server"):
            server = optarg
        elif opt == "--project-id":
            project_id = optarg
        elif opt == "--build-id":
            build_id = optarg

    # Try to determine reasonable defaults for Smoke options, if
    # necessary.
    if useSmoke:
        if project_id is None and "SMOKE_PROJECTID" not in os.environ:
            log.info("Falling back to Smoke's project id for 'komodo'.")
            project_id = smokeclient.getProjectId("komodo", server=server)
        if build_id is None and "SMOKE_BUILDID" not in os.environ:
            log.info("Falling back to a 'dev build id' for the local build.")
            versionTxt = os.path.join(os.path.dirname(__file__),
                                      os.pardir, "src", "version.txt")
            version = open(versionTxt, 'r').read()
            build_id = smokeclient.getDevBuildId("build-tree", version,
                                                 project_id=project_id,
                                                 server=server)

    _setUp()
    try:
        test(categories, modules, excludes=excludes, justList=justList,
             # Smoke options:
             useSmoke=useSmoke, server=server, project_id=project_id,
             build_id=build_id)
    finally:
        _tearDown()


if __name__ == "__main__":
    __file__ = sys.argv[0]
    sys.exit( main(sys.argv) )


